from typing import Dict, Any, Union, Type, List, Optional

import attr
from attr.exceptions import NotAnAttrsClassError

try:
    import dataclasses
except ModuleNotFoundError:
    dataclasses = None


@attr.attrs
class Field:
    name: str = attr.attrib()
    field_type: Type = attr.attrib()
    mandatory: bool = attr.attrib()
    validator: Optional[callable] = attr.attrib(default=None)
    converter: Optional[callable] = attr.attrib(default=None)


def resolve_types(to_resolve: Dict[Union[Type, str], Any], globals: Dict[str, Any]) -> Dict[Type, Any]:
    return {globals.get(k) if isinstance(k, str) else k: v for k, v in to_resolve.items()}


def get_fields(obj_type: Type) -> List[Field]:
    try:
        return [Field(f.name, f.type, f.default == attr.NOTHING, f.validator, f.converter)
                for f in attr.fields(obj_type)]
    except NotAnAttrsClassError:
        try:
            return [Field(f.name, f.type, f.default == dataclasses.MISSING and f.default_factory == dataclasses.MISSING)
                    for f in dataclasses.fields(obj_type)]
        except (TypeError, AttributeError):
            raise TypeError('can only serialize attrs or dataclass classes')


def normalize_method(method) -> callable:
    return method.__func__ if isinstance(method, staticmethod) else method



from yasoo.deserialization import Deserializer
from yasoo.serialization import Serializer

_default_serializer = Serializer()
serialize = _default_serializer.serialize
serializer = _default_serializer.register()
serializer_of = _default_serializer.register

_default_deserializer = Deserializer()
deserialize = _default_deserializer.deserialize
deserializer = _default_deserializer.register()
deserializer_of = _default_deserializer.register



import logging
from enum import Enum
from inspect import signature
from typing import Dict, Any, Union, Mapping, Iterable, Callable, Type, Optional

from yasoo.constants import ENUM_VALUE_KEY
from yasoo.utils import resolve_types, get_fields, normalize_method, Field

_logger = logging.getLogger(__name__)


class Serializer:
    def __init__(self) -> None:
        super().__init__()
        self._custom_serializers = {}

    def register(self, type_to_register: Optional[Type] = None):
        """
        Registers a custom serialization method that takes the object to be serialized and returns a json-serializable
        dictionary.

        :param type_to_register: The type of objects this method serializes. Can be a string, but then ``serialize``
            should be called with the ``globals`` parameter.
        """

        def registration_method(serialization_method: Union[Callable[[Any], Dict[str, Any]], staticmethod]):
            method = normalize_method(serialization_method)
            t = type_to_register
            if t is None:
                t = next(iter(signature(method).parameters.values())).annotation
            self._custom_serializers[t] = method
            return serialization_method

        return registration_method

    def serialize(self,
                  obj,
                  type_key: Optional[str] = '__type',
                  fully_qualified_types: bool = True,
                  globals: Optional[Dict[str, Any]] = None
                  ) -> Dict[str, Any]:
        """
        Serializes an object to a json-serializable dictionary.

        :param obj: The object to serialize.
        :param type_key: The key in the resulting dictionary to contain the type name for non-primitive objects.
            Can be ``None`` to omit this key and rely on type hints when deserializing.
        :param fully_qualified_types: Whether to use fully qualified type names in the ``type_key``,
            i.e. whether to write "my_package.MyType" or just "MyType".
        :param globals: If custom serialization methods were registered and used forward reference
            ('Foo' instead of Foo), this parameter should be a dictionary from type name to type, most easily
            acquired using the built-in ``globals()`` function.
        """
        if globals:
            self._custom_serializers = resolve_types(self._custom_serializers, globals)
        result = self._serialize(obj, type_key, fully_qualified_types, inner=False)
        return _convert_to_json_serializable(result)

    def _serialize(self, obj, type_key, fully_qualified_types, inner=True):
        serialization_method = self._custom_serializers.get(type(obj))
        if serialization_method:
            result = serialization_method(obj)
        else:
            try:
                fields = get_fields(type(obj))
                result = {f.name: self._serialize(getattr(obj, f.name), type_key, fully_qualified_types)
                          for f in fields}
                self._warn_for_possible_problems_in_deserialization(obj, fields, result)
            except TypeError:
                if isinstance(obj, Enum):
                    result = {ENUM_VALUE_KEY: obj.value}
                elif not inner:
                    raise
                else:
                    return obj

        if type_key is not None:
            type_value = '.'.join(
                (obj.__class__.__module__, obj.__class__.__name__)) if fully_qualified_types else obj.__class__.__name__
            result[type_key] = type_value
        return result

    @staticmethod
    def _warn_for_possible_problems_in_deserialization(obj, fields: Iterable[Field], data: Dict[str, Any]) -> None:
        for f in fields:
            if f.validator is not None and not isinstance(data[f.name], dict):
                try:
                    value = f.converter(data[f.name]) if f.converter is not None else data[f.name]
                except:
                    message = 'Field "{}" in obj "{}" has value {} that could not be converted using its converter'.format(
                        f.name, obj.__class__.__name__, data[f.name])
                    _logger.warning(message)
                    continue

                try:
                    f.validator(obj, f, value)
                except:
                    message = 'Field "{}" in obj "{}" has value {} that doesn\'t match this field\'s validator'.format(
                        f.name, obj.__class__.__name__, value)
                    _logger.warning(message)
                    continue
            if f.converter is not None:
                _logger.warning('Field "{}" in obj "{}" has a converter'.format(f.name, obj.__class__.__name__))


def _convert_to_json_serializable(obj) -> Union[int, float, str, list, dict, None]:
    if isinstance(obj, int) or isinstance(obj, float) or isinstance(obj, str) or obj is None:
        return obj
    if isinstance(obj, Mapping):
        return {key: _convert_to_json_serializable(value) for key, value in obj.items()}
    if isinstance(obj, Iterable):
        return [_convert_to_json_serializable(item) for item in obj]
    raise TypeError('Found object of type "{}" which cannot be serialized'.format(type(obj).__name__))



import json
from enum import Enum
from importlib import import_module
from inspect import signature
from typing import Optional, Type, Union, Callable, Dict, Any, TypeVar

from yasoo.constants import ENUM_VALUE_KEY
from yasoo.utils import resolve_types, get_fields, normalize_method

T = TypeVar('T')


class Deserializer:
    def __init__(self) -> None:
        super().__init__()
        self._custom_deserializers: Dict[Type[T], Callable[[Dict[str, Any]], T]] = {}

    def register(self, type_to_register: Optional[Union[Type, str]] = None):
        """
        Registers a custom deserialization method that takes a dictionary and returns an instance of the registered
        type.

        :param type_to_register: The type of objects this method deserializes to. Can be a string, but then
            ``deserialize`` should be called with the ``globals`` parameter.
        """
        def registration_method(deserialization_method: Union[Callable[[Dict[str, Any]], Any], staticmethod]):
            method = normalize_method(deserialization_method)
            t = type_to_register
            if t is None:
                t = signature(method).return_annotation
            self._custom_deserializers[t] = method
            return deserialization_method

        return registration_method

    def deserialize(self,
                    data: Dict[str, Any],
                    obj_type: Optional[Type[T]] = None,
                    type_key: Optional[str] = '__type',
                    globals: Optional[Dict[str, Any]] = None) -> T:
        """
        Deserializes an object from a dictionary.

        :param data: The dictionary.
        :param obj_type: The type of the object to deserialize. Can only be ``None`` if ``data`` contains a type key.
        :param type_key: The key in ``data`` that contains the type name for non-primitive objects.
            Can be ``None`` if this key was omitted during serialization and deserialization should rely on type hints.
        :param globals: If custom deserialization methods were registered and used forward reference
            ('Foo' instead of Foo), this parameter should be a dictionary from type name to type, most easily
            acquired using the built-in ``globals()`` function.
        """
        if globals:
            self._custom_deserializers = resolve_types(self._custom_deserializers, globals)

        return self._deserialize(data, obj_type, type_key, globals)

    def _deserialize(self,
                     data: Dict[str, Any],
                     obj_type: Optional[Type[T]],
                     type_key: Optional[str],
                     globals: Optional[Dict[str, Any]]):
        obj_type = self._get_object_type(obj_type, data, type_key, globals)
        if type_key in data:
            data.pop(type_key)

        deserialization_method = self._custom_deserializers.get(obj_type)
        if deserialization_method:
            return deserialization_method(data)

        if issubclass(obj_type, Enum):
            return obj_type(data[ENUM_VALUE_KEY])

        fields = {f.name: f for f in get_fields(obj_type)}
        self._check_for_missing_fields(data, fields, obj_type)
        self._check_for_extraneous_fields(data, fields, obj_type)
        self._load_inner_fields(data, fields, type_key, globals)
        return obj_type(**data)

    def _load_inner_fields(self, data, fields, type_key, globals):
        for key, value in data.items():
            if not isinstance(value, dict):
                continue
            field = fields[key]
            data[key] = self._deserialize(value, field.field_type, type_key, globals)

    @staticmethod
    def _check_for_missing_fields(data, fields, obj_type):
        missing = {name for name, field in fields.items() if name not in data and field.mandatory}
        if missing:
            raise ValueError('Missing fields "{}" for object type "{}". Data is:\n{}'.format('", "'.join(missing),
                                                                                             obj_type.__name__,
                                                                                             json.dumps(data)))

    @staticmethod
    def _check_for_extraneous_fields(data, fields, obj_type):
        extraneous = set(data.keys()).difference(fields)
        if extraneous:
            raise ValueError(
                'Found extraneous fields "{}" for object type "{}". Data is:\n{}'.format('", "'.join(extraneous),
                                                                                         obj_type.__name__,
                                                                                         json.dumps(data)))

    @staticmethod
    def _get_object_type(obj_type: Optional[Type[T]],
                         data: Dict[str, Any],
                         type_key: str,
                         globals: Optional[Dict[str, Any]]
                         ) -> Type:
        if type_key in data:
            return Deserializer._get_type(data[type_key], globals)
        if obj_type is None:
            raise ValueError(
                'type key not found in data and obj type could not be inferred.\nData: {}'.format(json.dumps(data)))
        return obj_type

    @staticmethod
    def _get_type(type_name: str, globals: Optional[Dict[str, Any]]) -> Type:
        if '.' not in type_name:
            return globals.get(type_name)
        module_name = type_name[:type_name.rindex('.')]
        class_name = type_name[len(module_name) + 1:]
        return getattr(import_module(module_name), class_name)
